@@ -1,3 +1,19 @@
+# <@LICENSE>
+# Copyright 2004 Apache Software Foundation
+# 
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+# 
+#     http://www.apache.org/licenses/LICENSE-2.0
+# 
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+# </@LICENSE>
+
 =head1 NAME
 
 Mail::SpamAssassin::PersistentAddrList - persistent address list base class
@@ -19,6 +35,8 @@ SpamAssassin will call:
 All persistent address list implementations, used by the auto-whitelist
 code to track known-good email addresses, use this as a base class.
 
+See C<Mail::SpamAssassin::DBBasedAddrList> for an example.
+
 =head1 METHODS
 
 =over 4
@@ -28,7 +46,12 @@ code to track known-good email addresses, use this as a base class.
 package Mail::SpamAssassin::PersistentAddrList;
 
 use strict;
-use vars qw{ @ISA };
+use warnings;
+use bytes;
+
+use vars qw{
+  @ISA
+};
 
 @ISA = qw();
 
@@ -60,7 +83,7 @@ SpamAssassin classes.
 
 sub new_checker {
   my ($factory, $main) = @_;
-  die "unimpled base method";	# override this
+  die "auto-whitelist: unimplemented base method";	# override this
 }
 
 ###########################################################################
@@ -88,7 +111,7 @@ a C<count> key and a C<totscore> key.
 sub get_addr_entry {
   my ($self, $addr) = @_;
   my $entry = { };
-  die "unimpled base method";	# override this
+  die "auto-whitelist: unimplemented base method";	# override this
   return $entry;
 }
 
@@ -103,7 +126,7 @@ given entry, and then return the new entry.
 
 sub add_score {
     my ($self, $entry, $score) = @_;
-    die "unimpled base method"; # override this
+    die "auto-whitelist: unimplemented base method"; # override this
 }
 
 ###########################################################################
@@ -116,7 +139,7 @@ This method should remove the given entry from the whitelist database.
 
 sub remove_entry {
   my ($self, $entry) = @_;
-  die "unimpled base method";	# override this
+  die "auto-whitelist: unimplemented base method";	# override this
 }
 
 ###########################################################################
@@ -135,3 +158,7 @@ sub finish {
 ###########################################################################
 
 1;
+
+=back
+
+=cut
