@@ -1,17 +1,35 @@
+# <@LICENSE>
+# Copyright 2004 Apache Software Foundation
+# 
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+# 
+#     http://www.apache.org/licenses/LICENSE-2.0
+# 
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+# </@LICENSE>
+
 =head1 NAME
 
 URIDNSBL - look up URLs against DNS blocklists
 
-This works by analysing message text and HTML for URLs, extracting the
-domain names from those, querying their NS records in DNS, resolving
-the hostnames used therein, and querying various DNS blocklists for
-those IP addresses.  This is quite effective.
-
 =head1 SYNOPSIS
 
   loadplugin    Mail::SpamAssassin::Plugin::URIDNSBL
   uridnsbl	URIBL_SBLXBL    sbl-xbl.spamhaus.org.   TXT
 
+=head1 DESCRIPTION
+
+This works by analysing message text and HTML for URLs, extracting the
+domain names from those, querying their NS records in DNS, resolving
+the hostnames used therein, and querying various DNS blocklists for
+those IP addresses.  This is quite effective.
+
 =head1 CONFIGURATION
 
 =over 4
@@ -37,10 +55,14 @@ C<lookuptype> is the type of lookup (B<TXT> or B<A>).   Note that you must also
 define a body-eval rule calling C<check_uridnsbl()> to use this.
 
 An RHSBL zone is one where the domain name is looked up, as a string; e.g. a
-URI using the domain C<foo.com> will cause a lookup of C<foo.com.uriblzone.net>.
-Note that hostnames are stripped from the domain used in the URIBL lookup,
-so the domain C<foo.bar.com> will look up C<bar.com.uriblzone.net>, and
-C<foo.bar.co.uk> will look up C<bar.co.uk.uriblzone.net>.
+URI using the domain C<foo.com> will cause a lookup of
+C<foo.com.uriblzone.net>.  Note that hostnames are stripped from the domain
+used in the URIBL lookup, so the domain C<foo.bar.com> will look up
+C<bar.com.uriblzone.net>, and C<foo.bar.co.uk> will look up
+C<bar.co.uk.uriblzone.net>.
+
+If a URI consists IP address instead of a hostname, the IP address is looked
+up (using the standard reversed quads method) in each C<rhsbl_zone>.
 
 Example:
 
@@ -89,8 +111,11 @@ not going to be listed in URIBLs.
 package Mail::SpamAssassin::Plugin::URIDNSBL;
 
 use Mail::SpamAssassin::Plugin;
+use Mail::SpamAssassin::Constants qw(:ip);
 use Mail::SpamAssassin::Util;
+use Mail::SpamAssassin::Logger;
 use strict;
+use warnings;
 use bytes;
 
 use vars qw(@ISA);
@@ -108,31 +133,13 @@ sub new {
   my $self = $class->SUPER::new($samain);
   bless ($self, $class);
 
-  # TODO: use infrastructure from Mail::SpamAssassin::Dns!
-  eval {
-    require Net::DNS;
-    require Net::DNS::Resolver; 
+  # this can be effectively global, at least in each process, safely
 
-    $self->{res} = Net::DNS::Resolver->new();
-  };
-
-  if ($@) {
-    dbg ("failed to load Net::DNS::Resolver: $@");
-  }
-
-  if ($self->{res}) {
-    $self->{res}->defnames(0);
-    $self->{res}->dnsrch(0);
-    $self->{res}->retrans(3);
-    $self->{res}->retry(1);
-  }
+  $self->{finished} = { };
 
   $self->register_eval_rule ("check_uridnsbl");
+  $self->set_config($samain->{conf});
 
-  # set default config settings
-  $samain->{conf}->{uridnsbl_timeout} =		3;
-  $samain->{conf}->{uridnsbl_max_domains} =	20;
-  $samain->{conf}->{uridnsbl_skip_domains} =	{};
   return $self;
 }
 
@@ -149,7 +156,7 @@ sub parsed_metadata {
   my ($self, $opts) = @_;
   my $scanner = $opts->{permsgstatus};
 
-  if (!($self->{res} && $scanner->is_dns_available())) {
+  if (!$scanner->is_dns_available()) {
     $self->{dns_not_available} = 1;
     return;
   }
@@ -166,10 +173,7 @@ sub parsed_metadata {
   $scanstate->{active_rules_rhsbl} = { };
   $scanstate->{active_rules_revipbl} = { };
   foreach my $rulename (keys %{$scanner->{conf}->{uridnsbls}}) {
-    # handle both body and head_evals for now
-    # head_evals was 3.0.0, body_evals is 3.0.1 and beyond
-    next unless ($scanner->{conf}->is_rule_active('body_evals',$rulename) ||
-    		 $scanner->{conf}->is_rule_active('head_evals',$rulename));
+    next unless ($scanner->{conf}->is_rule_active('body_evals',$rulename));
 
     my $rulecf = $scanstate->{scanner}->{conf}->{uridnsbls}->{$rulename};
     if ($rulecf->{is_rhsbl}) {
@@ -181,118 +185,210 @@ sub parsed_metadata {
 
   $self->setup ($scanstate);
 
+
   # get all domains in message
-  # TODO! we need a method that provides more metadata about where
-  # the URI was found so we can ignore hammy decoys.
-  my %domlist = ( );
-  foreach my $uri ($scanner->get_uri_list()) {
-    my $dom = Mail::SpamAssassin::Util::uri_to_domain($uri);
-    if ($dom) {
-      if (exists $scanner->{main}->{conf}->{uridnsbl_skip_domains}->{$dom}) {
-        dbg("URIDNSBL: found domain $dom in skip list");
+
+  # don't keep dereferencing this
+  my $skip_domains = $scanner->{main}->{conf}->{uridnsbl_skip_domains};
+
+  # list of arrays to use in order
+  my @uri_ordered = ();
+
+  # Generate the full list of html-parsed domains.
+  my $uris = $scanner->get_uri_detail_list();
+
+  # go from uri => info to uri_ordered
+  # 0: a
+  # 1: form
+  # 2: img
+  # 3: !a_empty
+  # 4: parsed
+  # 5: a_empty
+  while (my($uri, $info) = each %{$uris}) {
+    # we want to skip mailto: uris
+    next if ($uri =~ /^mailto:/);
+
+    # no domains were found via this uri, so skip
+    next unless ($info->{domains});
+
+    my $entry = 3;
+
+    if ($info->{types}->{a}) {
+      $entry = 5;
+
+      # determine a vs a_empty
+      foreach my $at (@{$info->{anchor_text}}) {
+        if (length $at) {
+	  $entry = 0;
+	  last;
+	}
       }
-      else {
-        $domlist{$dom} = 1;
+    }
+    elsif ($info->{types}->{form}) {
+      $entry = 1;
+    }
+    elsif ($info->{types}->{img}) {
+      $entry = 2;
+    }
+    elsif ($info->{types}->{parsed} && (keys %{$info->{types}} == 1)) {
+      $entry = 4;
+    }
+
+    # take the usable domains and add to the ordered list
+    foreach ( keys %{ $info->{domains} } ) {
+      if (exists $skip_domains->{$_}) {
+        dbg("uridnsbl: domain $_ in skip list");
+        next;
+      }
+      $uri_ordered[$entry]->{$_} = 1;
+    }
+  }
+
+  # at this point, @uri_ordered is an ordered array of uri hashes
+
+  my %domlist = ();
+  while (keys %domlist < $scanner->{main}->{conf}->{uridnsbl_max_domains} && @uri_ordered) {
+    my $array = shift @uri_ordered;
+    next unless $array;
+
+    # run through and find the new domains in this grouping
+    my @domains = grep(!$domlist{$_}, keys %{$array});
+    next unless @domains;
+
+    # the new domains are all useful, just add them in
+    if (keys(%domlist) + @domains <= $scanner->{main}->{conf}->{uridnsbl_max_domains}) {
+      foreach (@domains) {
+        $domlist{$_} = 1;
+      }
+    }
+    else {
+      # trim down to a limited number - pick randomly
+      my $i;
+      while (@domains && keys %domlist < $scanner->{main}->{conf}->{uridnsbl_max_domains}) {
+        my $r = int rand (scalar @domains);
+        $domlist{splice (@domains, $r, 1)} = 1;
       }
     }
   }
 
-  # trim down to a limited number - pick randomly
-  my $i;
-  my @longlist = keys %domlist;
-  my @shortlist = ();
-  for ($i = $scanner->{main}->{conf}->{uridnsbl_max_domains}; $i > 0; $i--) {
-    my $r = int rand (scalar @longlist);
-    push (@shortlist, splice (@longlist, $r, 1));
-    last if (scalar @longlist <= 0);
-  }
-
   # and query
-  dbg ("URIDNSBL: domains to query: ".join(' ',@shortlist));
-  foreach my $dom (@shortlist) {
+  dbg("uridnsbl: domains to query: ".join(' ',keys %domlist));
+  foreach my $dom (keys %domlist) {
     $self->query_domain ($scanstate, $dom);
   }
 
   return 1;
 }
 
-sub parse_config {
-  my ($self, $opts) = @_;
+sub set_config {
+  my($self, $conf) = @_;
+  my @cmds = ();
 
-  my $key = $opts->{key};
-  if ($key eq 'uridnsbl') {
-    if ($opts->{value} =~ /^(\S+)\s+(\S+)\s+(\S+)$/) {
-      my $rulename = $1;
-      my $zone = $2;
-      my $type = $3;
+  push(@cmds, {
+    setting => 'uridnsbl_timeout',
+    default => 3,
+    type => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
+  });
 
-      $opts->{conf}->{uridnsbls}->{$rulename} = {
-	zone => $zone, type => $type,
-        is_rhsbl => 0
-      };
-      $self->inhibit_further_callbacks(); return 1;
+  push(@cmds, {
+    setting => 'uridnsbl_max_domains',
+    default => 20,
+    type => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
+  });
+
+  push (@cmds, {
+    setting => 'uridnsbl',
+    code => sub {
+      my ($self, $key, $value, $line) = @_;
+      if ($value =~ /^(\S+)\s+(\S+)\s+(\S+)$/) {
+        my $rulename = $1;
+        my $zone = $2;
+        my $type = $3;
+        $self->{uridnsbls}->{$rulename} = {
+	  zone => $zone, type => $type,
+          is_rhsbl => 0
+        };
+      }
+      elsif ($value =~ /^$/) {
+        return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+      }
+      else {
+        return $Mail::SpamAssassin::Conf::INVALID_VALUE;
+      }
     }
-  }
-  elsif ($key eq 'urirhsbl') {
-    if ($opts->{value} =~ /^(\S+)\s+(\S+)\s+(\S+)$/) {
-      my $rulename = $1;
-      my $zone = $2;
-      my $type = $3;
+  });
 
-      $opts->{conf}->{uridnsbls}->{$rulename} = {
-	zone => $zone, type => $type,
-        is_rhsbl => 1
-      };
-      $self->inhibit_further_callbacks(); return 1;
+  push (@cmds, {
+    setting => 'urirhsbl',
+    code => sub {
+      my ($self, $key, $value, $line) = @_;
+      if ($value =~ /^(\S+)\s+(\S+)\s+(\S+)$/) {
+        my $rulename = $1;
+        my $zone = $2;
+        my $type = $3;
+        $self->{uridnsbls}->{$rulename} = {
+	  zone => $zone, type => $type,
+          is_rhsbl => 1
+        };
+      }
+      elsif ($value =~ /^$/) {
+        return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+      }
+      else {
+        return $Mail::SpamAssassin::Conf::INVALID_VALUE;
+      }
     }
-  }
-  elsif ($key eq 'urirhssub') {
-    if ($opts->{value} =~ /^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)$/) {
-      my $rulename = $1;
-      my $zone = $2;
-      my $type = $3;
-      my $subrule = $4;
+  });
 
-      $opts->{conf}->{uridnsbls}->{$rulename} = {
-	zone => $zone, type => $type,
-        is_rhsbl => 1, is_subrule => 1
-      };
-
-      $opts->{conf}->{uridnsbl_subs}->{$zone} ||= { };
-      $opts->{conf}->{uridnsbl_subs}->{$zone}->{$subrule} = {
-        rulename => $rulename
-      };
-
-      $self->inhibit_further_callbacks(); return 1;
+  push (@cmds, {
+    setting => 'urirhssub',
+    code => sub {
+      my ($self, $key, $value, $line) = @_;
+      if ($value =~ /^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)$/) {
+        my $rulename = $1;
+        my $zone = $2;
+        my $type = $3;
+        my $subrule = $4;
+        $self->{uridnsbls}->{$rulename} = {
+	  zone => $zone, type => $type,
+          is_rhsbl => 1, is_subrule => 1
+        };
+        $self->{uridnsbl_subs}->{$zone} ||= { };
+        $self->{uridnsbl_subs}->{$zone}->{$subrule} = {
+          rulename => $rulename
+        };
+      }
+      elsif ($value =~ /^$/) {
+        return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+      }
+      else {
+        return $Mail::SpamAssassin::Conf::INVALID_VALUE;
+      }
     }
-  }
-  elsif ($key eq 'uridnsbl_timeout') {
-    $opts->{conf}->{uridnsbl_timeout} = $opts->{value};
-    $self->inhibit_further_callbacks(); return 1;
-  }
-  elsif ($key eq 'uridnsbl_max_domains') {
-    $opts->{conf}->{uridnsbl_max_domains} = $opts->{value};
-    $self->inhibit_further_callbacks(); return 1;
-  }
-  elsif ($key eq 'uridnsbl_skip_domain') {
-    foreach my $domain (split(/\s+/, $opts->{value})) {
-      $opts->{conf}->{uridnsbl_skip_domains}->{lc $domain} = 1;
+  });
+
+  push (@cmds, {
+    setting => 'uridnsbl_skip_domain',
+    default => {},
+    code => sub {
+      my ($self, $key, $value, $line) = @_;
+      if ($value =~ /^$/) {
+        return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+      }
+      foreach my $domain (split(/\s+/, $value)) {
+        $self->{uridnsbl_skip_domains}->{lc $domain} = 1;
+      }
     }
-    $self->inhibit_further_callbacks(); return 1;
-  }
-  return 0;
+  });
+
+  $conf->{parser}->register_commands(\@cmds);
 }
 
 sub check_tick {
   my ($self, $opts) = @_;
 
   return if ($self->{dns_not_available});
-
-  # do a microscopic sleep to give other processes/the DNS server
-  # time to get at the CPU
-  select (undef, undef, undef, 0.01);
-
-  $self->complete_lookups($opts->{permsgstatus}->{uribl_scanstate});
+  $self->complete_lookups($opts->{permsgstatus}->{uribl_scanstate}, 0.3);
   return 1;
 }
 
@@ -305,14 +401,18 @@ sub check_post_dnsbl {
   my $scanstate = $scan->{uribl_scanstate};
 
   # try to complete a few more
-  if (!$self->complete_lookups($scanstate)) {
+  if (!$self->complete_lookups($scanstate, 0.1)) {
     my $secs_to_wait = $scan->{conf}->{uridnsbl_timeout};
-    dbg ("waiting $secs_to_wait seconds for URIDNSBL lookups to complete");
-    while ($secs_to_wait-- >= 0) {
-      last if ($self->complete_lookups($scanstate));
-      sleep 1;
+    if ($secs_to_wait < 0) { $secs_to_wait = 0; }
+    my $now = time;
+    my $deadline = $now + $secs_to_wait;
+    dbg("uridnsbl: waiting $secs_to_wait seconds for URIDNSBL lookups to complete");
+
+    while ($now < $deadline) {
+      last if ($self->complete_lookups($scanstate, $deadline - $now));
+      $now = time;
     }
-    dbg ("done waiting for URIDNSBL lookups to complete");
+    dbg("uridnsbl: done waiting for URIDNSBL lookups to complete");
   }
 
   foreach my $rulename (keys %{$scanstate->{active_rules_revipbl}},
@@ -346,31 +446,50 @@ sub setup {
 sub query_domain {
   my ($self, $scanstate, $dom) = @_;
 
-  #warn "domain $dom\n"; return;
+  #warn "uridnsbl: domain $dom\n";
+  #return;
 
   $dom = lc $dom;
-  return if $scanstate->{seen_domain}->{$dom}; $scanstate->{seen_domain}->{$dom}=1;
-  $self->log_dns_result ("querying domain $dom");
+  return if $scanstate->{seen_domain}->{$dom};
+  $scanstate->{seen_domain}->{$dom} = 1;
+  $self->log_dns_result("querying domain $dom");
 
   my $obj = {
     querystart => time,
     dom => $dom
   };
 
+  my $single_dnsbl = 0;
   if ($dom =~ /^\d+\.\d+\.\d+\.\d+$/) { 
-    $self->lookup_dnsbl_for_ip ($scanstate, $obj, $dom);
+    my $IPV4_ADDRESS = IPV4_ADDRESS;
+    my $IP_PRIVATE = IP_PRIVATE;
+    # only look up the IP if it is public and valid
+    if ($dom =~ /^$IPV4_ADDRESS$/ && $dom !~ /^$IP_PRIVATE$/) {
+      $self->lookup_dnsbl_for_ip($scanstate, $obj, $dom);
+      # and check the IP in RHSBLs too
+      if ($dom =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/) {
+	$dom = "$4.$3.$2.$1";
+	$single_dnsbl = 1;
+      }
+    }
   }
   else {
+    $single_dnsbl = 1;
+  }
+
+  if ($single_dnsbl) {
     # look up the domain in the RHSBL subset
     my $cf = $scanstate->{active_rules_rhsbl};
     foreach my $rulename (keys %{$cf}) {
       my $rulecf = $scanstate->{scanner}->{conf}->{uridnsbls}->{$rulename};
-      $self->lookup_single_dnsbl ($scanstate, $obj, $rulename,
-                          $dom, $rulecf->{zone}, $rulecf->{type});
+      $self->lookup_single_dnsbl($scanstate, $obj, $rulename,
+				 $dom, $rulecf->{zone}, $rulecf->{type});
     }
 
     # perform NS, A lookups to look up the domain in the non-RHSBL subset
-    $self->lookup_domain_ns ($scanstate, $obj, $dom);
+    if ($dom !~ /^\d+\.\d+\.\d+\.\d+$/) {
+      $self->lookup_domain_ns($scanstate, $obj, $dom);
+    }
   }
 }
 
@@ -383,7 +502,7 @@ sub lookup_domain_ns {
   return if $scanstate->{pending_lookups}->{$key};
 
   # dig $dom ns
-  my $ent = $self->start_lookup ($scanstate, 'NS', $self->{res}->bgsend ($dom, 'NS'));
+  my $ent = $self->start_lookup ($scanstate, 'NS', $self->res_bgsend($dom, 'NS'));
   $ent->{obj} = $obj;
   $scanstate->{pending_lookups}->{$key} = $ent;
 }
@@ -391,17 +510,30 @@ sub lookup_domain_ns {
 sub complete_ns_lookup {
   my ($self, $scanstate, $ent, $dom) = @_;
 
-  my $packet = $self->{res}->bgread($ent->{sock});
-  $self->close_ent_socket ($ent);
+  my $packet = $ent->{response_packet};
   my @answer = $packet->answer;
 
+  my $IPV4_ADDRESS = IPV4_ADDRESS;
+  my $IP_PRIVATE = IP_PRIVATE;
+
   foreach my $rr (@answer) {
     my $str = $rr->string;
     next unless (defined($str) && defined($dom));
     $self->log_dns_result ("NSs for $dom: $str");
 
     if ($str =~ /IN\s+NS\s+(\S+)/) {
-      $self->lookup_a_record($scanstate, $ent->{obj}, $1);
+      my $nsmatch = $1;
+
+      if ($nsmatch =~ /^\d+\.\d+\.\d+\.\d+\.?$/) {
+	$nsmatch =~ s/\.$//;
+	# only look up the IP if it is public and valid
+	if ($nsmatch =~ /^$IPV4_ADDRESS$/ && $nsmatch !~ /^$IP_PRIVATE$/) {
+	  $self->lookup_dnsbl_for_ip($scanstate, $ent->{obj}, $nsmatch);
+	}
+      }
+      else {
+	$self->lookup_a_record($scanstate, $ent->{obj}, $nsmatch);
+      }
     }
   }
 }
@@ -415,7 +547,7 @@ sub lookup_a_record {
   return if $scanstate->{pending_lookups}->{$key};
 
   # dig $hname a
-  my $ent = $self->start_lookup ($scanstate, 'A', $self->{res}->bgsend ($hname, 'A'));
+  my $ent = $self->start_lookup ($scanstate, 'A', $self->res_bgsend($hname, 'A'));
   $ent->{obj} = $obj;
   $scanstate->{pending_lookups}->{$key} = $ent;
 }
@@ -423,11 +555,7 @@ sub lookup_a_record {
 sub complete_a_lookup {
   my ($self, $scanstate, $ent, $hname) = @_;
 
-  my $packet = $self->{res}->bgread($ent->{sock});
-  $self->close_ent_socket ($ent);
-  my @answer = $packet->answer;
-
-  foreach my $rr (@answer) {
+  foreach my $rr ($ent->{response_packet}->answer) {
     my $str = $rr->string;
     $self->log_dns_result ("A for NS $hname: $str");
 
@@ -448,8 +576,8 @@ sub lookup_dnsbl_for_ip {
   my $cf = $scanstate->{active_rules_revipbl};
   foreach my $rulename (keys %{$cf}) {
     my $rulecf = $scanstate->{scanner}->{conf}->{uridnsbls}->{$rulename};
-    $self->lookup_single_dnsbl ($scanstate, $obj, $rulename,
-			$revip, $rulecf->{zone}, $rulecf->{type});
+    $self->lookup_single_dnsbl($scanstate, $obj, $rulename,
+			       $revip, $rulecf->{zone}, $rulecf->{type});
   }
 }
 
@@ -462,7 +590,7 @@ sub lookup_single_dnsbl {
 
   # dig $ip txt
   my $ent = $self->start_lookup ($scanstate, 'DNSBL',
-				$self->{res}->bgsend ($item, $qtype));
+        $self->res_bgsend($item, $qtype));
   $ent->{obj} = $obj;
   $ent->{rulename} = $rulename;
   $ent->{zone} = $dnsbl;
@@ -478,9 +606,12 @@ sub complete_dnsbl_lookup {
   my $rulename = $ent->{rulename};
   my $rulecf = $conf->{uridnsbls}->{$rulename};
 
-  my $packet = $self->{res}->bgread($ent->{sock});
-  $self->close_ent_socket ($ent);
+  my $packet = $ent->{response_packet};
   my @answer = $packet->answer;
+
+  my $uridnsbl_subs = $conf->{uridnsbl_subs}->{$ent->{zone}};
+  my $uridnsbl_subs_bits = 0;
+  $uridnsbl_subs_bits |= $_ for keys %{$uridnsbl_subs};
   foreach my $rr (@answer)
   {
     next if ($rr->type ne 'A' && $rr->type ne 'TXT');
@@ -490,23 +621,37 @@ sub complete_dnsbl_lookup {
 
     if (!$rulecf->{is_subrule}) {
       # this zone is a simple rule, not a set of subrules
-      $self->got_dnsbl_hit ($scanstate, $ent, $rdatastr, $dom, $rulename);
+      # skip any A record that isn't on 127/8
+      if ($rr->type eq 'A' && $rr->rdatastr !~ /^127\./) {
+	warn("uridnsbl: bogus rr for domain=$dom, rule=$rulename, id=" .
+            $packet->header->id." rr=".$rr->string);
+	next;
+      }
+      $self->got_dnsbl_hit($scanstate, $ent, $rdatastr, $dom, $rulename);
     }
     else {
-      my $uridnsbl_subs = $conf->{uridnsbl_subs}->{$ent->{zone}};
+      # skip any A record that isn't on 127/8 if we're not looking for
+      # any bits in the first octet, this is a workaround for bug 3997
+      if ($rr->type eq 'A' && $rr->rdatastr !~ /^127\./ &&
+	  !($uridnsbl_subs_bits & 0xff000000))
+      {
+	warn("uridnsbl: bogus rr: domain=$dom, zone=$ent->{zone}, id=" .
+            $packet->header->id." rr=".$rr->string);
+	next;
+      }
       foreach my $subtest (keys (%{$uridnsbl_subs}))
       {
         my $subrulename = $uridnsbl_subs->{$subtest}->{rulename};
 
         if ($subtest eq $rdatastr) {
-          $self->got_dnsbl_hit ($scanstate, $ent, $rdatastr, $dom, $subrulename);
+          $self->got_dnsbl_hit($scanstate, $ent, $rdatastr, $dom, $subrulename);
         }
         # bitmask
         elsif ($subtest =~ /^\d+$/) {
           if ($rdatastr =~ m/^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/ &&
               Mail::SpamAssassin::Util::my_inet_aton($rdatastr) & $subtest)
           {
-            $self->got_dnsbl_hit ($scanstate, $ent, $rdatastr, $dom, $subrulename);
+            $self->got_dnsbl_hit($scanstate, $ent, $rdatastr, $dom, $subrulename);
           }
         }
         # regular expression
@@ -524,7 +669,7 @@ sub got_dnsbl_hit {
   my ($self, $scanstate, $ent, $str, $dom, $rulename) = @_;
 
   $str =~ s/\s+/  /gs;	# long whitespace => short
-  dbg ("URIDNSBL: domain \"$dom\" listed ($rulename): $str");
+  dbg("uridnsbl: domain \"$dom\" listed ($rulename): $str");
 
   if (!defined $scanstate->{hits}->{$rulename}) {
     $scanstate->{hits}->{$rulename} = { };
@@ -535,10 +680,10 @@ sub got_dnsbl_hit {
 # ---------------------------------------------------------------------------
 
 sub start_lookup {
-  my ($self, $scanstate, $type, $sock) = @_;
+  my ($self, $scanstate, $type, $id) = @_;
   my $ent = {
     type => $type,
-    sock => $sock
+    id => $id
   };
   $scanstate->{queries_started}++;
   $ent;
@@ -550,7 +695,7 @@ sub start_lookup {
 # are, the next lookup in the sequence will be kicked off.
 
 sub complete_lookups {
-  my ($self, $scanstate) = @_;
+  my ($self, $scanstate, $timeout) = @_;
   my %typecount = ();
   my $stillwaiting = 0;
 
@@ -562,18 +707,23 @@ sub complete_lookups {
   $scanstate->{queries_started} = 0;
   $scanstate->{queries_completed} = 0;
 
+  my $nfound = $self->{main}->{resolver}->poll_responses($timeout);
+  $nfound ||= 'no';
+  dbg ("uridnsbl: select found $nfound socks ready");
+
   foreach my $key (keys %{$pending}) {
     my $ent = $pending->{$key};
-
     my $type = $ent->{type};
-    $key =~ /:(\S+)$/; my $val = $1;
 
-    if (!$self->{res}->bgisready ($ent->{sock})) {
+    if (!exists ($self->{finished}->{$ent->{id}})) {
       $typecount{$type}++;
       #$stillwaiting = 1;
       next;
     }
 
+    $ent->{response_packet} = delete $self->{finished}->{$ent->{id}};
+    $key =~ /:(\S+)$/; my $val = $1;
+
     if (LOG_COMPLETION_TIMES) {
       my $secs = (time - $ent->{start});
       my $totalsecs = (time - $ent->{obj}->{querystart});
@@ -590,7 +740,7 @@ sub complete_lookups {
     elsif ($type eq 'DNSBL') {
       $self->complete_dnsbl_lookup ($scanstate, $ent, $val);
       my $totalsecs = (time - $ent->{obj}->{querystart});
-      dbg ("URIDNSBL: query for ".$ent->{obj}->{dom}." took ".
+      dbg("uridnsbl: query for ".$ent->{obj}->{dom}." took ".
 		$totalsecs." seconds to look up ($val)");
     }
 
@@ -598,11 +748,11 @@ sub complete_lookups {
     delete $scanstate->{pending_lookups}->{$key};
   }
 
-  dbg ("URIDNSBL: queries completed: ".$scanstate->{queries_completed}.
+  dbg("uridnsbl: queries completed: ".$scanstate->{queries_completed}.
 		" started: ".$scanstate->{queries_started});
 
   if (1) {
-    dbg ("URIDNSBL: queries active: ".
+    dbg("uridnsbl: queries active: ".
 	join (' ', map { "$_=$typecount{$_}" } sort keys %typecount)." at ".
 	localtime(time));
   }
@@ -620,7 +770,7 @@ sub complete_lookups {
       if ($scanstate->{last_count} == $numkeys) {
 	$scanstate->{times_count_was_same}++;
 	if ($scanstate->{times_count_was_same} > 20) {
-	  dbg ("URIDNSBL: escaping: must have lost requests");
+	  dbg("uridnsbl: escaping: must have lost requests");
 	  $self->abort_remaining_lookups ($scanstate);
 	  $stillwaiting = 0;
 	}
@@ -644,34 +794,31 @@ sub abort_remaining_lookups  {
   foreach my $key (keys %{$pending})
   {
     if (!$foundone) {
-      dbg ("URIDNSBL: aborting remaining lookups");
+      dbg("uridnsbl: aborting remaining lookups");
       $foundone = 1;
     }
 
-    $self->close_ent_socket ($pending->{$key});
     delete $pending->{$key};
   }
 }
 
-sub close_ent_socket {
-  my ($ent) = @_;
-  if ($ent->{sock}) {
-    $ent->{sock}->close();
-    delete $ent->{sock};
-  }
-}
-
 # ---------------------------------------------------------------------------
 
+sub res_bgsend {
+  my ($self, $host, $type) = @_;
+
+  return $self->{main}->{resolver}->bgsend($host, $type, undef, sub {
+        my $pkt = shift;
+        my $id = shift;
+        $self->{finished}->{$id} = $pkt;
+      });
+}
+
 sub log_dns_result {
-  my $self = shift;
-  #Mail::SpamAssassin::dbg ("URIDNSBL: ".join (' ', @_));
+  #my $self = shift;
+  #Mail::SpamAssassin::dbg("uridnsbl: ".join (' ', @_));
 }
 
 # ---------------------------------------------------------------------------
 
-sub dbg { Mail::SpamAssassin::dbg (@_); }
-
-# ---------------------------------------------------------------------------
-
 1;
